from typing import Any

from amino import Map, List

from ribosome.machine.transition import may_handle
from ribosome.unite import UniteKind
from ribosome.unite.machine import UniteMachine, UniteTransitions
from ribosome.machine.state import Component, ComponentMachine
from ribosome.nvim import NvimFacade
from ribosome.machine.interface import MachineI

from myo.components.unite.message import UniteHistory, UniteCommands
from myo.components.unite.data import UniteNames, HistorySource, CommandsSource

history_source = HistorySource()
commands_source = CommandsSource()
sources = List(history_source, commands_source)
run_action = Map(name='run', handler=UniteNames.run, desc='run command')
delete_action = Map(name='delete', handler=UniteNames.delete, desc='delete command')
command_kind = UniteKind(UniteNames.command, List(run_action))
history_command_kind = UniteKind(UniteNames.history_command, List(run_action, delete_action))
kinds = List(command_kind, history_command_kind)


class MyoUniteTransitions(UniteTransitions, Component):

    def unite_cmd(self, cmd: str) -> None:
        args = ' '.join(self.msg.unite_args)
        self.vim.cmd('Unite {} {}'.format(cmd, args))

    @may_handle(UniteCommands)
    def commands(self) -> None:
        self.unite_cmd(UniteNames.commands)

    @may_handle(UniteHistory)
    def history(self) -> None:
        self.unite_cmd(UniteNames.history)


class Unite(UniteMachine, ComponentMachine):

    def __init__(self, vim: NvimFacade, title: str, parent: MachineI) -> None:
        UniteMachine.__init__(self)
        ComponentMachine.__init__(self, vim, MyoUniteTransitions, title, parent)

__all__ = ('Unite',)



from ribosome.machine.message_base import message, json_message

Run = json_message('Run', 'command')
RunLine = message('RunLine', 'target', varargs='args')
ShellRun = json_message('ShellRun', 'shell')
RunLatest = json_message('RunLatest')
RunTest = json_message('RunTest')
RunVimTest = json_message('RunVimTest')
RunIn = json_message('RunIn', 'target', 'line')
RunChained = message('RunChained', varargs='commands')
RebootCommand = json_message('RebootCommand', 'command')
Dispatch = json_message('Dispatch', 'command')
AddCommand = json_message('AddCommand', 'name')
AddShellCommand = json_message('AddShellCommand', 'name')
AddShell = json_message('AddShell', 'name')
AddVimCommand = json_message('AddVimCommand', 'name')
SetShellTarget = message('SetShellTarget', 'shell', 'target')
CommandAdded = json_message('CommandAdded', 'command')
CommandExecuted = message('CommandExecuted', 'job', 'log_path')
CommandShow = message('CommandShow')
CommandHistoryShow = message('CommandHistoryShow')
LoadHistory = message('LoadHistory')
StoreHistory = message('StoreHistory')
DeleteHistory = message('DeleteHistory', 'ident')



import abc

from amino import IO
from amino.lazy import lazy

from ribosome.nvim.components import Syntax
from ribosome.util.callback import VimCallback


class OutputSyntax(VimCallback):

    @lazy
    def syntax(self):
        return Syntax(self.target)

    def _line_re(self, index, rest='.*'):
        return '\%{}l{}'.format(index + 1, rest)

    def _whole_line_re(self, index, rest='.*'):
        return '^{}$'.format(self._line_re(index, rest))

    def _match(self, grp, rex, *a, **kw):
        return IO.call(self.syntax.match, grp, rex, *a, **kw)

    def _match_line(self, grp, index, *a, **kw):
        return self._match(grp, self._whole_line_re(index), *a, **kw)

    def _cont(self, grp, rex, cont):
        return self._match(grp, rex, 'contained', containedin=cont)

    def _link(self, left, right):
        return IO.call(self.syntax.link, left, right)

    def _hi(self, grp, *a, **kw):
        return IO.call(self.syntax.highlight, grp, *a, **kw)


class SimpleSyntax(OutputSyntax):

    def __call__(self, lines):
        return IO.call(self.syntax, lines)

    @abc.abstractmethod
    def syntax(self, lines):
        ...


class LineGroups(OutputSyntax):

    def __call__(self, lines):
        'transparent'
        return ((lines.with_index.map2(self._line)).sequence(IO) +
                self._highlight)

    def _line(self, index, line):
        def run(grp):
            prefixed = '{}{}'.format('Myo', grp)
            return self._match_line(prefixed, index)
        return line.syntax_group / run | IO.zero

    @property
    def _highlight(self):
        return (
            self._link('MyoError', 'Error') +
            self._link('MyoPath', 'Directory')
        )

__all__ = ('OutputSyntax', 'SimpleSyntax')



import os
import abc
import tempfile
from pathlib import Path
from typing import Tuple

from psutil import Process

from libtmux.pane import Pane as LTPane

from amino.io import io

from amino import __, List, Boolean, Maybe, _, Map, Just, Try, Either
from amino.lazy import lazy

from ribosome.record import optional_field, either_field, bool_field

from myo.ui.tmux.view import View
from myo.ui.tmux.adapter import Adapter
from myo.util import parse_int
from myo.ui.tmux.util import parse_window_id, parse_pane_id, parse_session_id


def child_pids(root: int) -> Either[Exception, List[int]]:
    return (
        Try(Process, root) /
        __.children() /
        List.wrap /
        __.map(_.pid)
    )


def descendant_pids(root: int) -> Either[Exception, List[int]]:
    def recurse(pids: List[int]) -> Either[Exception, List[int]]:
        return pids.traverse(descendant_pids, Either) / _.join / pids.add
    return child_pids(root) // recurse


class Pane(View):
    id = optional_field(int)
    pid = optional_field(int)
    shell_pid = optional_field(int)
    window_id = either_field(int)
    session_id = either_field(int)
    log_path = optional_field(Path)
    pin = bool_field()
    focus = bool_field()
    kill = bool_field()

    @property
    def id_s(self):
        return self.id / '%{}'.format

    @property
    def desc(self):
        id = self.id / ' %{}'.format | ''
        pid = self.pid / ' | pid -> {}'.format | ''
        pin = ' !' if self.pin else ''
        return 'P{} \'{}\'{}{} {}'.format(id, self.name, pin, pid,
                                          self.size_desc)

    def open(self, pa: 'PaneAdapter'):
        return self.set(id=pa.id_i.to_maybe, shell_pid=pa.pid.to_maybe,
                        session_id=pa.session_id_i, window_id=pa.window_id_i)

    @property
    def _str_extra(self):
        return self.id.to_list + super()._str_extra


class VimPane(Pane):
    pane_name = '<vim>'

    def __new__(cls, *a, name=None, **kw):
        n = Maybe(name) | VimPane.pane_name
        return super().__new__(cls, *a, name=n, **kw)

    @property
    def desc(self):
        return 'V{}'.format(super().desc)


class PaneI(metaclass=abc.ABCMeta):

    @abc.abstractproperty
    def _raw_pid(self) -> Maybe[int]:
        ...

    @lazy
    def pid(self) -> Maybe[int]:
        return self._raw_pid // parse_int

    @property
    def command_pid(self) -> Either[Exception, int]:
        return self.pid // child_pids // __.head.to_either('no child pids')

    @property
    def command_pids(self) -> Either[Exception, List[int]]:
        return self.pid // descendant_pids


class PaneData(PaneI):

    def __init__(self, data: Map) -> None:
        self.data = data

    def attr(self, name):
        return (
            self.data.get(name)
            .or_else(self.data.get('pane_{}'.format(name)))
        )

    @lazy
    def id(self):
        return self.attr('id')

    @property
    def id_i(self):
        return self.id // parse_pane_id

    @lazy
    def window_id(self):
        return self.attr('window_id')

    @property
    def window_id_i(self):
        return self.window_id // parse_window_id

    @lazy
    def session_id(self):
        return self.attr('session_id')

    @property
    def session_id_i(self):
        return self.session_id // parse_session_id

    @property
    def _raw_pid(self):
        return self.data.get('pane_pid')


class NativePane(LTPane):

    def __init__(self, window=None, **kwargs):
        if not window:
            raise ValueError('Pane must have ``Window`` object')
        self.window = window
        self.session = self.window.session
        self.server = self.session.server
        self._pane_id = kwargs['pane_id']
        self._data = Map(kwargs)

    @property
    def _info(self):
        return self._data


class PaneAdapter(Adapter, PaneI):

    @lazy
    def id(self):
        return self.native._pane_id

    @lazy
    def id_i(self):
        return parse_pane_id(self.id)

    @property
    def _raw_pid(self):
        return Try(lambda: self.native.pid)

    @io
    def resize(self, size, horizontal):
        f = __.set_width if horizontal else __.set_height
        return f(size)(self.native)

    @property  # type: ignore
    @io
    def kill(self):
        return self.cmd('kill-pane')

    def split(self, horizontal):
        return self.native.split_window(vertical=not horizontal)

    @lazy
    def width(self):
        return int(self.native.width)

    @lazy
    def height(self):
        return int(self.native.height)

    @lazy
    def size(self) -> Tuple[int, int]:
        return self.width, self.height

    @lazy
    def left(self):
        return int(self.native.left)

    @lazy
    def top(self):
        return int(self.native.top)

    @lazy
    def position(self):
        return self.left, self.top

    def send_keys(self, cmd, enter=True, suppress_history=True):
        return self.native.send_keys(cmd, enter=enter,
                                     suppress_history=suppress_history)

    @property
    def capture(self):
        return List.wrap(self.cmd('capture-pane', '-p'))

    @property
    def session_id(self):
        return self.native.session.id

    @lazy
    def session_id_i(self):
        return parse_session_id(self.session_id)

    @lazy
    def window_id(self):
        return self.native.window.id

    @lazy
    def window_id_i(self):
        return parse_window_id(self.window_id)

    def __repr__(self):
        return 'PA({}, {}, {})'.format(self.id, self.size, self.position)

    @property
    def running(self) -> Boolean:
        return self.command_pid.present

    @property
    def not_running(self) -> Boolean:
        return Boolean(not self.running)

    @property
    def _pipe_filter(self):
        # return 'cat'
        return 'sed -u -e \'s/\r//g\' -e \'s/\x1b\[[0-9;?]*[mlK]//g\''

    def pipe(self, base):
        uid = os.getuid()
        tmpdir = Path(tempfile.gettempdir()) / 'myo-{}'.format(uid) / base
        tmpdir.mkdir(exist_ok=True, parents=True)
        (fh, fname) = tempfile.mkstemp(prefix='pane-', dir=str(tmpdir))
        self.cmd('pipe-pane', '{} > {}'.format(self._pipe_filter, fname))
        return Just(Path(fname))

    def move_above(self, other: 'PaneAdapter'):
        self.cmd('move-pane', '-b', '-s', other.id)

    def move_below(self, other: 'PaneAdapter'):
        self.cmd('move-pane', '-s', other.id)

    def swap(self, other: 'PaneAdapter'):
        self.cmd('swap-pane', '-d', '-s', other.id)

    def __eq__(self, other):
        return isinstance(other, PaneAdapter) and self.id == other.id

    @property
    def active(self):
        return self.native.active == '1'

    def focus(self):
        self.native.select_pane()

    def run_command(self, line):
        self.quit_copy_mode()
        self.send_keys(line, suppress_history=False)

    @property
    def in_copy_mode(self):
        return self.native.in_mode == '1'

    def quit_copy_mode(self):
        if self.in_copy_mode:
            self.send_keys('C-c')

__all__ = ('Pane', 'VimPane', 'PaneAdapter')



import amino
from amino import __, List, Left, IO, Map, _, L

from ribosome.machine.transition import may_handle, Error, handle
from ribosome.machine.base import unit_nio
from ribosome.machine.transition import Fatal
from ribosome.machine.messages import RunIO, Stage1
from ribosome.machine.state import Component

from myo.components.core.dispatch import VimDispatcher
from myo.components.core.message import Initialized, ParseOutput
from myo.output import VimCompiler, CustomOutputHandler, Parsing
from myo.output.main import OutputHandler
from myo.command import CommandJob

error_no_output_events = 'no events in parse result'


class Core(Component):

    @may_handle(Stage1)
    def stage_1(self):
        return Initialized().pub.at(1)

    @may_handle(Initialized)
    def initialized(self):
        started = unit_nio(__.vars.set_p('started', True))
        return (self.data.set(initialized=True) + VimDispatcher(self.vim), started)

    @may_handle(Error)
    def error(self):
        m = self.msg.message
        handler = (L(self.log.caught_exception)('transition', _) if
                   amino.development and isinstance(m, Exception) else
                   self.log.error)
        handler(m)

    @handle(ParseOutput)
    def parse_output(self):
        opt = self.msg.options
        job = self._ensure_job(self.msg.job)
        def handle(parser):
            def display(result):
                return (
                    IO.now(Left(Fatal(error_no_output_events)))
                    if result.events.empty else
                    parser.display(result, opt)
                )
            return RunIO(parser.parse(self.msg.output, self.msg.path) // display)
        return self._error_handler(job) / handle

    def _error_handler(self, job):
        langs = self.msg.options.get('langs') / List.wrap | job.langs
        parser_spec = self.msg.options.get('parser').o(job.command.parser)
        def check(p):
            return (p if isinstance(p, OutputHandler) else
                    CustomOutputHandler(self.vim, p))
        return (
            self._inst_callbacks(parser_spec, None, self._special_parsers) /
            check
        ).o(self._langs_parsing(langs)).lmap(Fatal)

    @property
    def _special_parsers(self):
        return Map(compiler=VimCompiler)

    def _langs_parsing(self, langs):
        return langs.empty.no.either('command has no langs',
                                     Parsing(self.vim, langs))

    def _ensure_job(self, data):
        return (data if isinstance(data, CommandJob) else
                CommandJob(command=data))

__all__ = ('Core',)



import re
from uuid import UUID

from ribosome.machine.messages import Nop
from ribosome.machine.message_base import Message
from ribosome.machine.transition import may_handle, handle, Fatal
from ribosome.util.callback import parse_callback_spec
from ribosome.machine.base import UnitIO
from ribosome.record import decode_json, encode_json
from ribosome.machine.messages import Stage1
from ribosome.machine.state import Component

from amino import L, _, List, Try, __, Maybe, Map, I, IO, Just, Boolean, Left, Either
from myo.command import Command, VimCommand, ShellCommand, Shell, CommandJob, TransientCommandJob
from myo.util import amend_options, Ident
from myo.components.core.message import Parse, ParseOutput
from myo.components.command.message import (
    Run, ShellRun, Dispatch, AddCommand, AddShellCommand, AddShell, AddVimCommand, SetShellTarget, CommandExecuted,
    RunTest, RunVimTest, CommandAdded, CommandShow, RunLatest, LoadHistory, StoreHistory, RunLine, RunChained,
    RebootCommand, DeleteHistory, CommandHistoryShow
)
from myo.components.command.util import assemble_vim_test_line


class Reboot:

    def __init__(self, job) -> None:
        self.job = job

    @property
    def name(self):
        return self.job.name


class CommandComponent(Component):
    _test_cmd_name = '<test>'
    _test_cmd_var = 'create_test_command'

    def _add(self, tpe: type, **strict):
        return self._from_opt(tpe, **strict) / self._add_command

    def _add_command(self, cmd: Command):
        return List(
            self.data.cat(cmd), CommandAdded(cmd, options=self.msg.options)
        )

    @may_handle(Stage1)
    def stage_i(self):
        line = self.vim.vars(self._last_test_line_var) | ''
        shell = self.vim.vars(self._last_test_shell_var)
        test_cmd = self.vim.vars.pb(self._test_cmd_var) | Boolean(True)
        opt1 = Map(line=line)
        opt = shell / (lambda a: opt1.cat(('shell', a))) | opt1
        return (
            test_cmd.cata(AddShellCommand(name=self._test_cmd_name, options=opt), Nop()),
            LoadHistory()
        )

    @property
    def _history_var(self):
        return 'Myo_history'

    @property
    def _last_test_line_var(self):
        return 'Myo_last_test_line'

    @property
    def _last_test_shell_var(self):
        return 'Myo_last_test_shell'

    @handle(LoadHistory)
    def load_history(self):
        return (
            self.vim.vars.s(self._history_var) //
            L(decode_json)(_).to_either('failed to load history').lmap(Fatal) /
            self.data.commands.setter.history /
            self.data.setter.commands
        )

    @may_handle(StoreHistory)
    def store_history(self):
        return UnitIO(
            IO.delay(encode_json, self.data.commands.history).join_either /
            L(self).vim.vars.set(self._history_var, _)
        )

    @may_handle(DeleteHistory)
    def delete_history(self) -> Message:
        new = self.data.commands.delete_history(UUID(self.msg.ident))
        return self.data.set(commands=new)

    @handle(AddCommand)
    def add_command(self):
        return self._add(Command, name=self.msg.name)

    @may_handle(AddVimCommand)
    def add_vim_command(self):
        return self._add(VimCommand, name=self.msg.name)

    @handle(AddShellCommand)
    def add_shell_command(self):
        return self._add(ShellCommand, name=self.msg.name)

    @handle(AddShell)
    def add_shell(self):
        return self._from_opt(Shell, name=self.msg.name) / (
            lambda shell: (
                self._add_command(shell) +
                (shell.target / L(SetShellTarget)(shell, _) / _.pub).to_list
            )
        )

    @may_handle(CommandAdded)
    def command_added(self):
        if self.msg.options.get('start').exists(I):
            return Run(self.msg.command.uuid)

    @handle(Run)
    def run(self):
        return (
            self._command_job_fatal(self.msg.command) /
            L(Dispatch)(_, self.msg.options)
        )

    @may_handle(RunLine)
    def run_line(self):
        return Dispatch(ShellCommand(name='run_line',
                                     line=self.msg.args.join_tokens),
                        Map(target=self.msg.target))

    @handle(ShellRun)
    def run_in_shell(self) -> Either[Fatal, Message]:
        line = self.msg.options.get('line') | ''
        def send(cmd):
            return (
                self.data.transient_command(cmd),
                Dispatch(cmd, self.msg.options)
            )
        # double Maybe is necessary - the outer indicates success, the
        # inner is the type of the ShellCommand field
        return (
            self._shell_fatal(self.msg.shell) /
            _.ident /
            Maybe /
            (lambda a: ShellCommand(name='shell_run', shell=a, line=line, transient=True)) /
            send
        )

    @handle(RunLatest)
    def run_latest(self):
        return self.latest_job_fatal / Dispatch

    @handle(RunChained)
    def run_chained(self):
        special_rex = re.compile('s:(.*)')
        def special(name):
            return (
                (
                    self._vim_test_line //
                    L(self._test_line_job)(Map(), _) /
                    List.wrap //
                    _.head
                ).to_either('failed to create test line')
                if name == 'vimtest' else
                Left('invalid special command: {}'.format(name))
            )
        def resolve(name):
            return (
                Maybe(special_rex.match(name)).cata(
                    L(special)(__.group(1)),
                    lambda: (self.data.command(name) /
                             CommandJob.from_attr('command'))
                )
            )
        def cons(line, job):
            cmd = job.command.set(eval=False)
            return (
                Dispatch(
                    TransientCommandJob(prefix='chained', command=cmd,
                                        override_line=line),
                    Map()
                )
            )
        commands = (
            List.wrap(self.msg.commands) / resolve
        ).sequence(Either)
        lines = ((commands.eff() / __.resolved_line(self.vim)).value //
                 __.sequence(Either))
        chained = (
            (self._callback('chainer') & lines)
            .map2(lambda a, b: a(b))
            .to_either('chainer failed')
        )
        return (
            (chained & (commands // __.head.to_either('commands empty')))
            .map2(cons)
            .lmap(Fatal)
        )

    @handle(RebootCommand)
    def reboot_command(self):
        return (
            self._command_job_fatal(self.msg.command) /
            Reboot /
            L(Dispatch)(_, self.msg.options)
        )

    @handle(Dispatch)
    def dispatch(self):
        cmd = self.msg.command
        self.vim.vars.set_p('last_command', dict(name=cmd.name))
        self.vim.pautocmd('RunCommand')
        return self.data.dispatch_message(cmd, self.msg.options) / _.pub

    @may_handle(CommandExecuted)
    def command_executed(self):
        job = self.msg.job
        def set_log(path, cmd_lens):
            def run(cmd):
                self.log.debug('setting {} log to {}'.format(cmd.name, path))
                return cmd.set(log_path=Just(path))
            return cmd_lens.modify(run)
        data = (
            (self.msg.log_path.to_either('no log path') &
             self.data.command_lens(job.main_key))
            .map2(set_log) |
            self.data
        )
        if job.history:
            new = data.commands.add_history(job)
            return data.set(commands=new), StoreHistory()

    @handle(Parse)
    def parse(self):
        def find_log(c):
            return ((self.data.command(c.main_key) // _.log_path)
                    .to_either('{} has no log'.format(c)))
        def parse(c, l):
            self.log.debug('parsing {} from {}'.format(c, l))
            return (
                Try(l.read_text) /
                List.lines /
                L(ParseOutput)(c, _, l, self.msg.options) /
                _.pub
            )
        cmd = (
            self.msg.options.get('command')
            .cata(
                L(self.data.command_or_job)(_),
                lambda: self.latest_job_fatal / Left
            ) / __.left_or_map(lambda a: CommandJob(command=a))
        )
        log_path = cmd // find_log
        return (cmd & log_path).flat_map2(parse).lmap(Fatal)

    @handle(RunTest)
    def run_test(self):
        opt = self.msg.options
        return (
            opt.get('ctor')
            .to_either('no test ctor specified') //
            self._assemble //
            L(self._test_line_dispatch)(opt - 'ctor', _)
        ).to_either('failed to setup test line').lmap(Fatal)

    @handle(RunVimTest)
    def run_vim_test(self):
        return (
            self._vim_test_line //
            L(self._test_line_dispatch)(self.msg.options, _)
        ).lmap(Fatal)

    @may_handle(CommandShow)
    def show(self):
        msg = self.data.commands.commands / _.desc
        self.log.info(msg.join_lines)

    @may_handle(CommandHistoryShow)
    def history_show(self):
        msg = self.data.commands.history / _.desc
        self.log.info(msg.join_lines)

    def _command_job_fatal(self, ident):
        return (self.data.command_or_job(self.msg.command)
                .lmap(Fatal) /
                __.left_or_map(CommandJob.from_attr('command')))

    def _shell_fatal(self, ident: Ident) -> Either[str, Shell]:
        return self.data.shell(ident).lmap(Fatal)

    @property
    def latest_job_fatal(self):
        return self.data.commands.latest_command.lmap(Fatal)

    def _assemble(self, ctor):
        return parse_callback_spec(ctor) // __(self.vim)

    @property
    def _vim_test_line(self):
        return assemble_vim_test_line(self.vim)

    def _test_line_dispatch(self, options, line):
        return self._test_line_job(options, line).map2(Dispatch)

    def _test_line_job(self, options, line):
        def dispatch(cmd, line, langs, opt):
            shell = opt.get('shell')
            job = TransientCommandJob(prefix='test_line', command=cmd,
                                      override_line=line, override_langs=langs,
                                      override_shell=shell)
            return job, opt
        return self._test_line_params(options, line).map4(dispatch)

    def _test_line_params(self, options, line):
        glangs = self.vim.buffer.vars.pl('test_langs')
        langs = options.get('langs').or_else(glangs) | List()
        shell = self.vim.buffer.pvar_or_global('test_shell')
        target = self.vim.buffer.pvar_or_global('test_target')
        opt = amend_options(options, 'shell', shell)
        opt2 = amend_options(opt, 'target', target) + ('langs', langs)
        self.vim.vars.set(self._last_test_line_var, line)
        shell / L(self.vim.vars.set)('Myo_last_test_shell', _)
        return (
            self.data.command(self._test_cmd_name)
            .to_either('no test command') /
            (lambda a: (a, line, langs, opt2))
        )

__all__ = ('CommandComponent',)



from myo.output.reifier.base import Reifier as ReifierBase
from myo.output.data import OutputLine, EmptyLine
from myo.output.parser.python import FileEntry, PyErrorEntry

from amino import List, Just, __


class Reifier(ReifierBase):

    def _format_file(self, entry: FileEntry):
        fun = entry.fun / ' {}'.format | ''
        return '{}  {}{}'.format(str(self._truncate(entry.path)), entry.line, fun)

    def _format_error(self, entry):
        return '{} {}'.format(entry.exc, entry.error)

    def _format_code(self, entry):
        return entry.code

    def _file(self, event, entry: FileEntry):
        code = entry.code / __.format_lines(Just(entry), self._format_code) | List() / __.set(indent=2)
        return entry.format_lines(Just(entry), self._format_file) + code

    def _error(self, event, entry: PyErrorEntry):
        return (entry.format_lines(Just(event), self._format_error) +
                List(EmptyLine.create(event)))

    def _event(self, event):
        def dispatch(entry):
            return (
                self._file(event, entry)
                if isinstance(entry, FileEntry) else
                self._error(event, entry)
                if isinstance(entry, PyErrorEntry) else
                List()
            )
        return event.entries // dispatch

    def __call__(self, result) -> List[OutputLine]:
        return result.events // self._event

__all__ = ('Reifier',)



import neovim

from amino import List, L, _, I, Map

from ribosome import msg_command, json_msg_command, AutoPlugin
from ribosome.machine.scratch import Mapping
from ribosome.request.function import msg_function
from ribosome.machine.state import UpdateRecord
from ribosome.unite import mk_unite_candidates, mk_unite_action
from ribosome.unite.plugin import unite_plugin
from ribosome.settings import PluginSettings, Config, RequestHandler

from myo.logging import Logging
from myo.components.command.message import (AddVimCommand, Run, AddShellCommand, AddShell, ShellRun, RunTest, RunVimTest,
                                         CommandShow, RunLatest, RunLine, RunChained, RebootCommand, DeleteHistory,
                                         CommandHistoryShow)
from myo.components.tmux.message import (TmuxCreatePane, TmuxCreateSession, TmuxCreateLayout, TmuxSpawnSession, TmuxInfo,
                                      TmuxClosePane, TmuxPack, TmuxMinimize, TmuxRestore, TmuxToggle, TmuxFocus,
                                      TmuxOpenOrToggle, TmuxKill)
from myo.components.tmux.message import TmuxOpen
from myo.components.core.message import Parse, Resized
from myo.components.unite.message import UniteHistory, UniteCommands
from myo.components.unite.main import UniteNames, Unite
from myo.components.unite.format import unite_format
from myo.output.machine import EventPrev, EventNext
from myo.env import Env
from myo.components.core.main import Core
from myo.components.command.main import CommandComponent
from myo.components.tmux.main import Tmux

unite_candidates = mk_unite_candidates(UniteNames)
unite_action = mk_unite_action(UniteNames)


config = Config(
    name='myo',
    prefix='myo',
    state_type=Env,
    components=Map(core=Core, command=CommandComponent, tmux=Tmux, unite=Unite),
    settings=PluginSettings(),
    request_handlers=List(
        # RequestHandler.msg_cmd(Msg)('msg', prefix=Plain, sync=True)
    ),
    core_components=List('core'),
    default_components=List('command', 'tmux', 'unite'),
)


@unite_plugin('myo')
class MyoNvimPlugin(AutoPlugin, Logging):

    @json_msg_command(AddVimCommand)
    def myo_vim_command(self):
        pass

    @json_msg_command(AddShellCommand)
    def myo_shell_command(self):
        pass

    @json_msg_command(AddShell)
    def myo_shell(self):
        pass

    @json_msg_command(TmuxCreateSession)
    def myo_tmux_create_session(self):
        pass

    @msg_command(TmuxSpawnSession)
    def myo_tmux_spawn_session(self):
        pass

    @json_msg_command(TmuxCreateLayout)
    def myo_tmux_create_layout(self):
        pass

    @json_msg_command(TmuxCreatePane)
    def myo_tmux_create_pane(self):
        pass

    @json_msg_command(TmuxOpen)
    def myo_tmux_open(self):
        pass

    @msg_command(TmuxClosePane)
    def myo_tmux_close_pane(self):
        pass

    @msg_command(TmuxPack)
    def myo_tmux_pack(self):
        pass

    @json_msg_command(TmuxMinimize)
    def myo_tmux_minimize(self):
        pass

    @json_msg_command(TmuxRestore)
    def myo_tmux_restore(self):
        pass

    @json_msg_command(TmuxToggle)
    def myo_tmux_toggle(self):
        pass

    @msg_command(TmuxOpenOrToggle)
    def myo_tmux_open_or_toggle(self):
        pass

    @msg_command(TmuxFocus)
    def myo_tmux_focus(self):
        pass

    @json_msg_command(TmuxKill)
    def myo_tmux_kill(self):
        pass

    @json_msg_command(Run)
    def myo_run(self):
        pass

    @msg_command(RunLine)
    def myo_run_line(self):
        pass

    @json_msg_command(ShellRun)
    def myo_run_in_shell(self):
        pass

    @json_msg_command(RunLatest)
    def myo_run_latest(self):
        pass

    @msg_command(RunChained)
    def myo_run_chained(self):
        pass

    @json_msg_command(RebootCommand)
    def myo_reboot_command(self):
        pass

    @msg_command(TmuxInfo)
    def myo_tmux_show(self):
        pass

    @msg_command(CommandShow)
    def myo_command_show(self):
        pass

    @msg_command(CommandHistoryShow)
    def myo_command_history_show(self):
        pass

    @json_msg_command(Parse)
    def myo_parse(self):
        pass

    @msg_command(EventPrev)
    def myo_event_prev(self):
        pass

    @msg_command(EventNext)
    def myo_event_next(self):
        pass

    @json_msg_command(RunTest)
    def myo_test(self):
        pass

    @json_msg_command(RunVimTest)
    def myo_vim_test(self):
        pass

    @json_msg_command(UpdateRecord)
    def myo_update(self):
        pass

    def _eval(self, args, go):
        result = List.wrap(args).head.to_either('expression needed') // go
        return result.cata(self.log.error, I)

    @neovim.function('MyoEval', sync=True)
    def myo_eval(self, args):
        return self._eval(args, self.myo.eval_expr)

    @neovim.function('MyoTmuxEval', sync=True)
    def myo_tmux_eval(self, args):
        def mod(data, components):
            d = data.sub_states.get('tmux') | data
            p = components.get('tmux') | components
            return d, p
        return self._eval(args, L(self.myo.eval_expr)(_, mod))

    @neovim.autocmd('VimResized', sync=False)
    def vim_resized(self):
        if self.root is not None:
            self.root.send(Resized())

    @msg_function(Mapping)
    def myo_mapping(self):
        pass

    @msg_command(UniteHistory)
    def myo_unite_history(self):
        pass

    @msg_command(UniteCommands)
    def myo_unite_commands(self):
        pass

    @unite_candidates('history')
    def myo_unite_history_candidates(self, args):
        return self.myo.data.commands.history / unite_format

    @unite_candidates('commands')
    def myo_unite_commands_candidates(self, args):
        return self.myo.data.commands.commands / unite_format

    @unite_action('run', 'name')
    def myo_unite_run(self, ident):
        return Run(ident, options=Map())

    @unite_action('delete', 'ident')
    def myo_unite_delete(self, ident):
        return DeleteHistory(ident)

__all__ = ('MyoNvimPlugin',)



import time

from libtmux.common import TmuxRelationalObject

from amino import List

from myo.logging import Logging


class Adapter(Logging):

    def __init__(self, native) -> None:
        if not isinstance(native, TmuxRelationalObject):
            raise Exception(
                'passed {!r} to {}'.format(native, self.__class__.__name__))
        self.native = native

    def __repr__(self):
        return 'A({})'.format(self.native)

    def cmd(self, *a):
        start = time.time()
        c = self.native.cmd(*a)
        if c.stderr:
            self.log.error(List.wrap(c.stderr).join_lines)
        self.log.ddebug(
            'tmux cmd `{}` took {:.4f}'.format(' '.join(a),
                                               time.time() - start))
        return c.stdout

    def cmd_async(self, *a):
        self.native.cmd(*a)

__all__ = ('Adapter',)



from amino import List, Just, Lists

from ribosome import NvimFacade

from myo.env import Env
from myo.state import MyoState


class Myo(MyoState):

    def __init__(self, vim: NvimFacade, components: List[str]=List()) -> None:
        core = 'myo.components.core'
        super().__init__(vim, Lists.wrap(components).cons(core))

    @property
    def init(self) -> Env:
        return Env(vim_facade=Just(self.vim))

__all__ = ('Myo',)



from typing import Callable

from amino import List, __, _

from lenses import lens

from ribosome.record import field, list_field, bool_field

from myo.record import Record
from myo.ui.tmux.session import Session, VimSession
from myo.ui.tmux.window import VimWindow, Window
from myo.ui.tmux.layout import VimLayout, Layout
from myo.ui.tmux.pane import VimPane
from myo.util import Ident
from myo.components.tmux.watcher import Watcher

_is_vim_session = lambda a: isinstance(a, VimSession)
_is_vim_window = lambda a: isinstance(a, VimWindow)
_is_vim_layout = lambda a: isinstance(a, VimLayout)
_is_vim_pane = lambda a: isinstance(a, VimPane)


class TmuxState(Record):
    initialized = bool_field()
    sessions = list_field()
    instance_id = field(str, initial='', factory=lambda a: a if a else List.random_string(5))
    watcher = field(Watcher)

    @property
    def _str_extra(self):
        return self.sessions

    def session_lens(self, pred: Callable[[Session], bool]):
        return self.attr_lens_pred(_.sessions, pred) / __.bind(self)

    @property
    def vim_session(self):
        return self.sessions.find(_is_vim_session).to_either('no vim session')

    @property
    def vim_session_lens(self):
        return (self.sessions.find_lens_pred(_is_vim_session) /
                lens(self).sessions.add_lens)

    def window_lens(self, pred: Callable[[Window], bool]):
        sub = __.attr_lens_pred(_.windows, pred)
        return (
            (self.attr_lens(_.sessions, sub) / __.bind(self))
            .to_either('no such window: {}'.format(pred))
        )

    def window_lens_ident(self, ident: Ident):
        return self.window_lens(__.has_ident(ident))

    @property
    def vim_window(self):
        return (
            (self.vim_session // __.windows.find(_is_vim_window))
            .to_either('no vim window')
        )

    @property
    def vim_window_lens(self):
        return self.window_lens(_is_vim_window)

    def layout_lens(self, pred: Callable[[Layout], bool]):
        lay_sub = __.root.layout_lens(pred).map(lens().root.add_lens)
        win_sub = __.attr_lens(_.windows, lay_sub)
        return self.attr_lens(_.sessions, win_sub) / __.bind(self)

    def layout_lens_ident(self, ident: Ident):
        return self.layout_lens(__.has_ident(ident))

    @property
    def vim_layout(self):
        return self.vim_window // __.root.layouts.find(_is_vim_layout)

    @property
    def vim_layout_lens(self):
        return self.layout_lens(_is_vim_layout)

    def pane_lens(self, pred: Callable[[Layout], bool]):
        lay_sub = __.root.pane_lens(pred).map(lens().root.add_lens)
        win_sub = __.attr_lens(_.windows, lay_sub)
        return self.attr_lens(_.sessions, win_sub) / __.bind(self)

    def pane_lens_ident(self, ident: Ident):
        return self.pane_lens(__.has_ident(ident))

    @property
    def vim_pane(self):
        return self.vim_layout / _.panes / _.head

    @property
    def all_windows(self):
        return self.sessions // _.windows

    def window(self, ident: Ident):
        return self.all_windows.find(__.has_ident(ident))

    @property
    def all_panes(self):
        return self.all_windows // _.root.all_panes

    def pane(self, ident):
        return self.all_panes.find(__.has_ident(ident))

    @property
    def possibly_open_panes(self):
        return (
            self.all_panes
            .filter(_.id.present)
            .filter(lambda a: not _is_vim_pane(a))
        )

__all__ = ('TmuxState',)
